require "csv"

WEKA_PATH = "/usr/share/java/weka/weka.jar"
SEED = 1
TUNING_PERCENTAGE = 0.1
$random = Random.new(SEED)

class MLTechnique
    def initialize(name)
        @classifier = name
        @parameters = {}
        @filter = nil
    end
    
    def filter(filter)
        @filter = filter
    end
    
    def tune_max(dataset, instances, tuning_matrix, &block)
        return evaluate_configs(dataset, instances, tuning_matrix, &block).max_by { |e| e[1] }
    end
    
    def tune_min(dataset, instances, tuning_matrix, &block)
        return evaluate_configs(dataset, instances, tuning_matrix, &block).min_by { |e| e[1] }
    end
    
    def each_config
    end
    
    def params
        @parameters.clear
    end
    
    protected
    def param(key, value)
        @parameters[key] = value
    end
    
    def get_results(dataset, folds)
        #puts "java -cp \"#{WEKA_PATH}\" weka.classifiers.#@classifier -t \"#{dataset}\" -x #{folds} -s #{SEED} #{parameter_string} 2> /dev/null"
        basic_params_string = "-t \"#{dataset}\" -x #{folds} -s #{SEED}"
        ml_name = "weka.classifiers.#@classifier"
        unless @filter
            return `java -cp "#{WEKA_PATH}" #{ml_name} #{basic_params_string} #{parameter_string} 2> /dev/null`
        else
            ml_string = "#{ml_name} -- #{parameter_string}".gsub("\"", "\\\"")
            filter_string = @filter
            puts "java -cp \"#{WEKA_PATH}\" weka.classifiers.meta.FilteredClassifier #{basic_params_string} -F \"#{filter_string}\" -W #{ml_string} 2> /dev/null"
            return `java -cp "#{WEKA_PATH}" weka.classifiers.meta.FilteredClassifier #{basic_params_string} -F "#{filter_string}" -W #{ml_string} 2> /dev/null`
        end
    end
    
    def evaluate_configs(dataset, percentage, tuning_matrix)
        sample_path = dataset.sub(".csv", ".sample.csv")
        rows = []
        header = nil
        
        CSV.parse(File.read(dataset), headers: true, col_sep: ",") do |row|
            header = row.headers unless header
            rows << row
        end
        instances = (percentage*rows.size).to_i
        
        unless FileTest.exist?(sample_path)
            sample = rows.sample(instances, random: $random)
            CSV.open(sample_path, "w") do |csv|
                csv << header
                sample.each do |row|
                    csv << row
                end
            end
        end
        
        configs = []
        done = 0
        each_config do |config|
            performance = tuning_matrix[config.to_s]
            performance = performance.to_f if performance
            unless performance
                params *config
                performance = evaluate(sample_path, instances)
            end
            
            yield config, performance if block_given?
            
            configs.push [config.to_s, performance]
            done += 1            
        end
        
        return configs
    end
    
    private
    def parameter_string
        result = ""
        
        @parameters.each do |k, v|
            if v != true
                result += "#{k} #{v} "
            else
                result += "#{k} "
            end
        end
        
        return result
    end
end

class Regressor < MLTechnique
    def tune(dataset, instances, tuning_matrix={}, &block)
        tune_min(dataset, instances, tuning_matrix, &block)
    end
    
    def evaluate(dataset, folds)
        result = get_results(dataset, folds)
        
        return result.scan(/Mean absolute error\s*([0-9.]+)/).flatten.last.to_f
    end
end

class Classifier < MLTechnique
    def tune(dataset, instances, tuning_matrix={}, &block)
        tune_max(dataset, instances, tuning_matrix, &block)
    end
    
    def evaluate(dataset, folds)
        filtered = dataset.sub(".csv", "filtered.arff")
        
        unless FileTest.exist? filtered
            `java -cp "/usr/share/java/weka/weka.jar" weka.filters.unsupervised.attribute.NumericToNominal -R last -i "#{dataset}" -o "#{filtered}" 2> /dev/null`
        end
        
        result = get_results(filtered, folds)
        result_matches = result.scan(/Weighted Avg.\s*\-?[?0-9.]+\s*\-?[?0-9.]+\s*\-?[?0-9.]+\s*\-?[?0-9.]+\s*\-?[?0-9.]+\s*\-?[?0-9.]+\s*(\-?[0-9.]+)/).flatten
        auc = result_matches.last.to_f
        
        if auc < 0 || auc > 1
            puts result
            puts "Severe error: AUC cannot be < 0 or > 1. #{auc} instead."
            p result_matches
            exit
        end
        
        return auc
    end
end

class MultilayerPerceptronReg < Regressor
    def initialize
        super "functions.MultilayerPerceptron"
    end
    
    def params(hidden_layers, learning_rate, momentum, lr_decay, iterations)
        super()
        param "-H", hidden_layers
        param "-L", learning_rate
        param "-M", momentum
        param "-D", true if lr_decay
        param "-N", iterations
    end
    
    def each_config
        hl_space    = (1..5).to_a.map { |e| e * 10.0 }
        lr_space    = (1..9).to_a.map { |e| e / 10.0 }
        m_space     = (1..5).to_a.map { |e| e / 10.0 }
        lrd_space   = [true]
        i_space     = [500]
        
        hl_space.each do |hl|
            lr_space.each do |lr|
                m_space.each do |m|
                    lrd_space.each do |lrd|
                        i_space.each do |i|
                            yield [hl, lr, m, lrd, i]
                        end
                    end
                end
            end
        end
    end
end

class MultilayerPerceptronCl < Classifier
    def initialize
        super "functions.MultilayerPerceptron"
    end
    
    def params(hidden_layers, learning_rate, momentum, lr_decay, iterations)
        super()
        param "-H", hidden_layers
        param "-L", learning_rate
        param "-M", momentum
        param "-D", true if lr_decay
        param "-N", iterations
    end
    
    def each_config
        hl_space    = (1..5).to_a.map { |e| e * 10.0 }
        lr_space    = (1..9).to_a.map { |e| e / 10.0 }
        m_space     = (1..5).to_a.map { |e| e / 10.0 }
        lrd_space   = [true]
        i_space     = [500]
        
        hl_space.each do |hl|
            lr_space.each do |lr|
                m_space.each do |m|
                    lrd_space.each do |lrd|
                        i_space.each do |i|
                            yield [hl, lr, m, lrd, i]
                        end
                    end
                end
            end
        end
    end
end

class KNNReg < Regressor
    def initialize
        super "lazy.IBk"
    end
    
    def params(k, weighting, distance)
        super()
        param "-K", k
        
        param "-W", 0
        param "-I", true if weighting == 1
        param "-F", true if weighting == 2
        
        case distance
        when 0
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.ChebyshevDistance -R first-last\\\"\""
        when 1
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.EuclideanDistance -R first-last\\\"\""
        when 2
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.ManhattanDistance -R first-last\\\"\""
        when 3
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.MinkowskiDistance -P 2.0 -R first-last\\\"\""
        end
    end
    
    def each_config
        k_space     = (1..5).to_a
        w_space     = (0..2).to_a
        d_space     = (0..3).to_a
        win_space   = [0]
        
        k_space.each do |k|
            w_space.each do |w|
                d_space.each do |d|
                    next if k < 2 && w != 0
                    yield [k, w, d]
                end
            end
        end
    end
end

class KNNCl < Classifier
    def initialize
        super "lazy.IBk"
    end
    
    def params(k, weighting, distance)
        super()
        param "-K", k
        
        param "-W", 0
        param "-I", true if weighting == 1
        param "-F", true if weighting == 2
        
        case distance
        when 0
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.ChebyshevDistance -R first-last\\\"\""
        when 1
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.EuclideanDistance -R first-last\\\"\""
        when 2
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.ManhattanDistance -R first-last\\\"\""
        when 3
            param "-A", "\"weka.core.neighboursearch.LinearNNSearch -A \\\"weka.core.MinkowskiDistance -P 2.0 -R first-last\\\"\""
        end
    end
    
    def each_config
        k_space     = (1..5).to_a
        w_space     = (0..2).to_a
        d_space     = (0..3).to_a
        win_space   = [0]
        
        k_space.each do |k|
            w_space.each do |w|
                d_space.each do |d|
                    next if k < 2 && w != 0
                    yield [k, w, d]
                end
            end
        end
    end
end

class SMOReg < Regressor
    def initialize
        super "functions.SMOreg"
    end
    
    def params(kernel, exp, c)
        #weka.classifiers.functions.SMOreg -C 1.0 -N 0 -I "weka.classifiers.functions.supportVector.RegSMOImproved -T 0.001 -V -P 1.0E-12 -L 0.001 -W 1" -K "weka.classifiers.functions.supportVector.PolyKernel -E 1.0 -C 250007"
        super()
        param "-C", c.to_f
        param "-N", 0
        param "-I", "\"weka.classifiers.functions.supportVector.RegSMOImproved -T 0.001 -V -P 1.0E-12 -L 0.001 -W 1\""
        param "-K", "\"weka.classifiers.functions.supportVector.#{kernel} -E #{exp} -C 250007\""
    end
    
    def each_config
        k_space     = ["NormalizedPolyKernel", "PolyKernel"]
        e_space     = (2..5).to_a
        c_space     = (1..10).to_a
        
        k_space.each do |k|
            e_space.each do |e|
                c_space.each do |c|
                    yield [k, e, c]
                end
            end
        end
    end
end

class SMOCl < Classifier
    def initialize
        super "functions.SMO"
    end
    
    def params(kernel, exp, c)
        #weka.classifiers.functions.SMOreg -C 1.0 -N 0 -I "weka.classifiers.functions.supportVector.RegSMOImproved -T 0.001 -V -P 1.0E-12 -L 0.001 -W 1" -K "weka.classifiers.functions.supportVector.PolyKernel -E 1.0 -C 250007"
        super()
        param "-C", c.to_f
        param "-L", 0.001
        param "-P", "1.0E-12"
        param "-N", 0
        param "-V", -1
        param "-W", 1
        param "-calibrator", "\"weka.classifiers.functions.Logistic -R 1.0E-8 -M -1 -num-decimal-places 4\""
        param "-K", "\"weka.classifiers.functions.supportVector.#{kernel} -E #{exp} -C 250007\""
    end
    
    def each_config
        k_space     = ["NormalizedPolyKernel", "PolyKernel"]
        e_space     = (2..5).to_a
        c_space     = (1..10).to_a
        
        k_space.each do |k|
            e_space.each do |e|
                c_space.each do |c|
                    yield [k, e, c]
                end
            end
        end
    end
end

class RandomForestReg < Regressor
    def initialize
        super "trees.RandomForest"
    end
    
    def params(features)
        super()
        param "-P", 100
        param "-I", 100
        param "-num-slots", 1
        param "-K", features
        param "-S", SEED
        param "-V", 0.001
        param "-M", 1
    end
    
    def each_config
        f_space     = (1..10).to_a
        
        f_space.each do |f|
            yield [f]
        end
    end
end

class RandomForestCl < Classifier
    def initialize
        super "trees.RandomForest"
    end
    
    def params(features)
        super()
        param "-P", 100
        param "-I", 100
        param "-num-slots", 1
        param "-K", features
        param "-S", SEED
        param "-V", 0.001
        param "-M", 1
    end
    
    def each_config
        f_space     = (1..10).to_a
        
        f_space.each do |f|
            yield [f]
        end
    end
end

def tune_single(file, klass, dataset, percentage)
    cache_file = File.basename(dataset).sub(".csv", ".#{klass.name}.cache.txt")
    
    current_matrix = {}
    if FileTest.exist? cache_file
        data = File.read(cache_file).split("\n").map { |e| e.split(" => ") }
        data.each do |config, performance|
            current_matrix[config] = performance
        end
    end
    
    data = nil
    File.open(cache_file, "w") do |f| 
        data = klass.new.tune(dataset, percentage, current_matrix) do |config, performance|
            puts "\t\t" + config.to_s + " => "+ performance.to_s
            f.puts(config.to_s + " => " + performance.to_s)
        end
    end
    
    file.puts klass.name
    file.puts "\t" + data[0].to_s
    file.puts "\tBest value: " + data[1].to_s
end

def tune_all(dataset, todo)
    warn "Tuning for #{dataset}"
    fname = File.basename(dataset).sub(".csv", ".tuned.txt")
    
    mode = "w"
    if FileTest.exist? fname
        mode = "a+"
        content = File.read(fname).split("\n").select { |e| !e.start_with?("\t") }
        todo = todo.select { |c| !content.include?(c.name)}
    end
    
    File.open(fname, mode) do |f|
        todo.each do |c|
            warn "\tTuning #{c.name}"
            tune_single f, c, dataset, TUNING_PERCENTAGE
        end
    end
end

def tune_all_regressors(dataset)    
    tune_all dataset, [RandomForestReg, MultilayerPerceptronReg, SMOReg, KNNReg]
end

def tune_all_classifiers(dataset)
    tune_all dataset, [RandomForestCl, MultilayerPerceptronCl, SMOCl, KNNCl]
end

tune_all_regressors "results/dataset_tnpu.csv"
tune_all_regressors "results/dataset_tau.csv"
tune_all_regressors "results/dataset_au.csv"

tune_all_classifiers "results/dataset_bd.csv"
tune_all_classifiers "results/dataset_pbu.csv"
tune_all_classifiers "results/dataset_abu.csv"
